package jp.riken.neurodata.tools;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Base64;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.ImageOutputStream;

import loci.common.DebugTools;
import loci.formats.FormatException;
import loci.formats.ImageReader;
import loci.formats.gui.AWTImageTools;
import loci.formats.gui.BufferedImageReader;

public class BioFormatsImageThumbnail {

    protected static final float JPEG_QUALITY = 0.85f;
    protected static final int BACKGROUND_COLOR = 0x000000;
    protected static final double MAXIMUM_SCALE = 2.0;

    public static BufferedImage resize(final BufferedImage image, final int width, final int height) {
        final boolean hasAlpha = image.getColorModel().hasAlpha();
        final int imageWidth = image.getWidth();
        final int imageHeight = image.getHeight();
        final double scaleWidth = (double) width / (double) imageWidth;
        final double scaleHeight = (double) height / (double) imageHeight;
        double scale = Math.min(scaleWidth, scaleHeight);
        if (scale > MAXIMUM_SCALE) {
            // limit scaling size to maximum scale
            scale = MAXIMUM_SCALE;
        }
        int resizeWidth = imageWidth;
        int resizeHeight = imageHeight;
        Image resizeImage = image;
        if (scale != 1.0) {
            // resize image if dimension is different with requested dimension.
            resizeWidth = (int) (scale * imageWidth);
            resizeHeight = (int) (scale * imageHeight);
            resizeImage = image.getScaledInstance(resizeWidth, resizeHeight, Image.SCALE_AREA_AVERAGING);
        }
        final int type = hasAlpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;
        final BufferedImage ret = new BufferedImage(width, height, type);
        final Graphics2D g = ret.createGraphics();
        final Color c = new Color(image.getRGB(0, 0), hasAlpha);
        final int offsetX = (width - resizeWidth) / 2;
        final int offsetY = (height - resizeHeight) / 2;
        g.setColor(c);
        g.fillRect(0, 0, width, height);
        g.drawImage(resizeImage, offsetX, offsetY, resizeWidth, resizeHeight, null);
        g.dispose();

        return ret;
    }

    protected static BufferedImage removeAlphaChannel(final BufferedImage image, final int matColor) {
        if (!image.getColorModel().hasAlpha()) {
            return image;
        }
        final BufferedImage ret = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_RGB);
        final Graphics2D g = ret.createGraphics();
        g.setColor(new Color(matColor, false));
        g.fillRect(0, 0, image.getWidth(), image.getHeight());
        g.drawImage(image, 0, 0, null);
        g.dispose();

        return ret;
    }

    protected static byte[] getJpegByteArray(final BufferedImage image, final float quality, final int matColor)
            throws BioFormatsImageException {
        byte[] ret = null;
        try (final ByteArrayOutputStream bos = new ByteArrayOutputStream();
                final ImageOutputStream ios = ImageIO.createImageOutputStream(bos);) {
            final ImageWriter writer = ImageIO.getImageWritersByFormatName("jpeg").next();
            final ImageWriteParam param = writer.getDefaultWriteParam();
            if (param.canWriteCompressed()) {
                param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
                param.setCompressionQuality(quality);
            }
            writer.setOutput(ios);
            writer.write(null, new IIOImage(removeAlphaChannel(image, matColor), null, null), param);
            writer.dispose();
            ret = bos.toByteArray();
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    protected static byte[] getPngByteArray(final BufferedImage image) throws BioFormatsImageException {
        byte[] ret = null;
        try (final ByteArrayOutputStream bos = new ByteArrayOutputStream();
                final ImageOutputStream ios = ImageIO.createImageOutputStream(bos);) {
            final ImageWriter writer = ImageIO.getImageWritersByFormatName("png").next();
            final ImageWriteParam param = writer.getDefaultWriteParam();
            if (param.canWriteCompressed()) {
                param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
                param.setCompressionQuality(0.0f);
            }
            writer.setOutput(ios);
            writer.write(null, new IIOImage(image, null, null), param);
            writer.dispose();
            ret = bos.toByteArray();
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    public static String getDataUri(final BufferedImage image, final String mimeType) throws BioFormatsImageException {
        byte[] bytes = null;
        switch (mimeType) {
            case "image/png":
                bytes = getPngByteArray(image);
                break;
            case "image/jpeg":
                bytes = getJpegByteArray(image, JPEG_QUALITY, BACKGROUND_COLOR);
                break;
            default:
                throw new BioFormatsImageException("Unsupported image format: " + mimeType);
        }
        if (bytes.length == 0) {
            throw new BioFormatsImageException("Failed to convert to " + mimeType + " format");
        }

        return "data:" + mimeType + ";base64," + Base64.getEncoder().encodeToString(bytes);
    }

    public static BufferedImage getImageFromBioFormatsFile(final String path) throws BioFormatsImageException {
        BufferedImage ret = null;
        try (final BufferedImageReader reader = new BufferedImageReader(new ImageReader())) {
            reader.setFlattenedResolutions(false);
            reader.setNormalized(true);
            reader.setOriginalMetadataPopulated(false);
            reader.setMetadataFiltered(true);
            reader.setId(path);
            int series = 0;
            final int seriesCount = reader.getSeriesCount();
            for (int i = 0; i < seriesCount; i++) {
                reader.setSeries(i);
                if (reader.isThumbnailSeries()) {
                    series = i;
                    break;
                }
            }
            reader.setSeries(series);
            final int z = reader.getSizeZ() / 2; // middle of slices
            final int t = reader.getSizeT() / 2; // middle of frames
            final int c = 0; // first channel
            final int no = reader.getIndex(z, c, t);

            ret = reader.openThumbImage(no);
            final boolean is16bit = reader.getBitsPerPixel() > 8;
            final boolean isGrayScale = ret.getSampleModel().getNumBands() == 1;
            if (is16bit || isGrayScale) {
                // perform auto scaling if 16bit or gray scale thumbnail image.
                ret = AWTImageTools.autoscale(ret);
            }
        } catch (final OutOfMemoryError e) {
            throw new BioFormatsImageException("out of memory");
        } catch (final FormatException | IOException e) {
            throw new BioFormatsImageException("unsupported file: " + path);
        } catch (final Exception e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    public static BufferedImage getImageFromFile(final String path) throws BioFormatsImageException {
        BufferedImage ret = null;
        final File file = new File(path);
        try {
            ret = ImageIO.read(file);
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }

        return ret;
    }

    public static void saveFile(final String dataUri, final String path) throws BioFormatsImageException {
        final Pattern pattern = Pattern.compile("^data:(image/[^;]+);(base64|utf8),(.*)$", Pattern.MULTILINE);
        final Matcher matcher = pattern.matcher(dataUri);
        if (!matcher.find()) {
            throw new BioFormatsImageException("Unsupported data URI format");
        }
        final byte[] bytes = "utf8".equals(matcher.group(2)) ? matcher.group(3).getBytes()
                : Base64.getDecoder().decode(matcher.group(3));
        try (final FileOutputStream fos = new FileOutputStream(path);) {
            fos.write(bytes);
        } catch (final IOException e) {
            throw new BioFormatsImageException(e);
        }
    }

    public static boolean readThumbnail(final String path, final Map<String, Object> thumbnail) {
        boolean ret = true;
        try {
            final BufferedImage image = getImageFromBioFormatsFile(path);
            final int width = image.getWidth();
            final int height = image.getHeight();
            final String mimeType = image.getColorModel().hasAlpha() ? "image/png" : "image/jpeg";
            final String dataUri = getDataUri(image, mimeType);
            thumbnail.put("width", width);
            thumbnail.put("height", height);
            thumbnail.put("mimeType", mimeType);
            thumbnail.put("url", dataUri);
        } catch (final Exception e) {
            // ignore
            ret = false;
        }

        return ret;
    }

    public static void main(final String[] args) {
        if (args.length != 1) {
            System.out.println("filename argument required");
            System.exit(1);
        }
        final String path = args[0];
        try {
            DebugTools.enableLogging("OFF");
            BufferedImage image = getImageFromBioFormatsFile(path);
            final String format = image.getColorModel().hasAlpha() ? "png" : "jpeg";
            final String dataUri = getDataUri(image, "image/" + format);
            System.out.println(dataUri);
            saveFile(dataUri, "/var/tmp/thumbnail." + format);
        } catch (final Exception e) {
            e.printStackTrace();
            System.out.println(e.getMessage());
            System.exit(1);
        }
    }
}
